import os
import sys


class View(object):

    def __init__(self):
        self.show_verbose = False

    def info(self, message):
        print(message)

    def verbose(self, message):
        if self.show_verbose:
            print(message)

    def error(self, message):
        sys.stderr.write(message)
        sys.stderr.write(os.linesep)



import os
import json
import sys

import dew
from dew.buildoptions import BuildOptionsCache, BuildOptions
from dew.dewfile import DewFileParser
from dew.exceptions import DewError
from dew.projectprocessor import ProjectProcessor
from dew.storage import StorageController
from dew.view import View


def main() -> int:
    parser = dew.make_argparser()
    args = parser.parse_args()

    view = View()
    view.show_verbose = True if args.v else False

    storage = StorageController(os.getcwd())
    storage.ensure_directories_exist()

    options_cache = BuildOptionsCache(storage)
    options = options_cache.get_options()
    if not options:
        options = BuildOptions()

    apply_argument_options(options, args)
    options_cache.save_options(options)

    invalid_options = options.get_invalid_options()
    if len(invalid_options) > 0:
        view.error('The following options are either unset or set to invalid values:\n{0}\n'
                   'Please specify the correct values as arguments.'.format('\n'.join(invalid_options)))
        return 64

    dewfile_parser = DewFileParser()
    with open('dewfile.json') as file:
        dewfile_data = json.load(file)
    dewfile_parser.set_data(dewfile_data)
    dewfile = dewfile_parser.parse()

    project_processor = ProjectProcessor(storage, options, view)
    project_processor.set_data(dewfile)

    try:
        project_processor.process()
    except DewError:
        view.error('Dew did not finish successfully :(')
        return 1

    return 0


def apply_argument_options(options: BuildOptions, args) -> None:
    if args.cmake_generator:
        options.cmake_generator = args.cmake_generator


sys.exit(main())



import git.repo


def clone_repo(url, destination_dir, ref=None):
    kwargs = {
        'depth': 1
    }
    if ref:
        kwargs['b'] = ref
    try:
        repo = git.repo.Repo(destination_dir)
        repo.remotes[0].pull()
    except (git.NoSuchPathError, git.InvalidGitRepositoryError):
        git.repo.Repo.clone_from(url, destination_dir, **kwargs)
        return



import argparse


def make_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('--cmake-generator', dest='cmake_generator')
    parser.add_argument('-v', help='Be verbose', action='store_true')
    return parser



import os.path


class StorageController(object):

    def __init__(self, path):
        self.path = path

    def ensure_directories_exist(self):
        os.makedirs(self.get_storage_dir(), exist_ok=True)
        os.makedirs(self.get_sources_dir(), exist_ok=True)
        os.makedirs(self.get_builds_dir(), exist_ok=True)
        os.makedirs(self.get_downloads_dir(), exist_ok=True)
        os.makedirs(self.get_install_dir(), exist_ok=True)

    def get_storage_dir(self):
        return self.join_storage_dir_path()

    def get_sources_dir(self):
        return self.join_storage_dir_path('sources')

    def get_builds_dir(self):
        return self.join_storage_dir_path('builds')

    def get_downloads_dir(self):
        return self.join_storage_dir_path('downloads')

    def get_install_dir(self):
        return self.join_storage_dir_path('install')

    def get_storage_dir_name(self):
        return '.dew'

    def join_storage_dir_path(self, *args):
        return os.path.join(self.path, self.get_storage_dir_name(), *args)



from dew.buildoptions import BuildOptions
from dew.dependencygraph import DependencyGraph
from dew.dependencyprocessor import DependencyProcessor
from dew.dewfile import DewFile
from dew.storage import StorageController
from dew.view import View


class ProjectProcessor(object):

    def __init__(self, storage: StorageController, options: BuildOptions, view: View):
        self.storage = storage
        self.root_dewfile = None
        self.options = options
        self.view = view

    def set_data(self, dewfile: DewFile):
        self.root_dewfile = dewfile

    def process(self):
        dewfile_stack = [(self.root_dewfile, None)]

        # Dependency processors by name
        dependency_processors = {}

        graph = DependencyGraph()

        while len(dewfile_stack) > 0:
            dewfile, parent_name = dewfile_stack.pop()
            for dep in dewfile.dependencies:
                graph.add_dependency(dep.name, parent_name)

                dep_processor = DependencyProcessor(self.storage, self.view)
                dependency_processors[dep.name] = dep_processor

                dep_processor.set_data(dep, dewfile, self.options)

                self.view.info('Pulling dependency {0}...'.format(dep.name))
                dep_processor.pull()

                if dep_processor.has_dewfile():
                    dewfile_stack.append((dep_processor.get_dewfile(), dep.name))

        names_in_order = graph.resolve()

        for name in names_in_order:
            dep_processor = dependency_processors[name]
            self.view.info('Building dependency {0}...'.format(dep_processor.dependency.name))
            dep_processor.build()



import json
import os.path
from typing import List

from dew.storage import StorageController


class BuildOptions(object):
    def __init__(self):
        self.cmake_generator = None

    def get_invalid_options(self) -> List[str]:
        invalid_options = []

        if not self.cmake_generator:
            invalid_options.append('cmake_generator')

        return invalid_options


class BuildOptionsCache(object):
    def __init__(self, storage: StorageController):
        self.storage = storage

    def get_options(self) -> BuildOptions or None:
        cache_path = self.get_cache_file_path()
        if os.path.isfile(cache_path):
            with open(cache_path) as f:
                data = json.load(f)
                return self.from_data(data)
        else:
            return None

    def save_options(self, options: BuildOptions) -> None:
        cache_path = self.get_cache_file_path()
        with open(cache_path, 'w') as f:
            json.dump(self.to_dict(options), f, indent=4)

    def from_data(self, data) -> BuildOptions:
        options = BuildOptions()
        options.cmake_generator = data['cmake_generator']
        return options

    def to_dict(self, options: BuildOptions) -> dict:
        data = dict()
        data['cmake_generator'] = options.cmake_generator
        return data

    def get_cache_file_path(self) -> str:
        return os.path.join(self.storage.get_storage_dir(), 'cache.json')





class DewFile(object):
    def __init__(self):
        self.dependencies = []


class Dependency(object):
    def __init__(self):
        self.name = ""
        self.url = "",
        self.type = ""
        self.ref = ""


class DewFileParser(object):
    def __init__(self):
        self.data = None

    def set_data(self, data) -> None:
        self.data = data

    def parse(self) -> DewFile:
        dependencies = []
        dependencies_obj = self.data['dependencies']
        for dep in dependencies_obj:
            dependencies.append(self.parse_dependency(dep))

        dewfile = DewFile()
        dewfile.dependencies = dependencies
        return dewfile

    def parse_dependency(self, obj) -> Dependency:
        dep = Dependency()
        dep.name = obj['name']
        dep.url = obj['url']
        dep.type = obj['type']
        dep.ref = obj['ref']
        return dep



import json
import os.path
import subprocess
import shutil
from enum import Enum, auto

from dew.buildoptions import BuildOptions
from dew.dewfile import Dependency, DewFile, DewFileParser
from dew.exceptions import BuildError, PullError
from dew.storage import StorageController
from dew import git
from dew.view import View


class BuildSystem(Enum):
    UNKNOWN = auto()
    CMAKE = auto()
    MAKEFILE = auto()


class DependencyProcessor(object):
    def __init__(self, storage: StorageController, view: View):
        self.storage = storage
        self.dependency = None
        self.dewfile = None
        self.options = None
        self.view = view

    def set_data(self, dependency: Dependency, dewfile: DewFile, options: BuildOptions):
        self.dependency = dependency
        self.dewfile = dewfile
        self.options = options

    def process(self):
        self.pull()
        self.build()

    def pull(self):
        type = self.dependency.type
        if type == 'git':
            self.pull_git()
        elif type == 'local':
            self.pull_local()
        else:
            self.view.error('Cannot pull, unknown dependency type {0}'.format(type))
            raise PullError()

    def build(self):
        buildsystem = self.get_buildsystem()
        if buildsystem == BuildSystem.MAKEFILE:
            self.build_makefile()
        elif buildsystem == BuildSystem.CMAKE:
            self.build_cmake()
        else:
            self.view.error('Cannot build, unkown build system')
            raise BuildError()

    def has_dewfile(self) -> bool:
        dewfile_path = os.path.join(self.get_src_dir(), 'dewfile.json')
        return os.path.isfile(dewfile_path)

    def get_dewfile(self) -> DewFile or None:
        dewfile_path = os.path.join(self.get_src_dir(), 'dewfile.json')
        if os.path.isfile(dewfile_path):
            parser = DewFileParser()
            with open(dewfile_path) as f:
                data = json.load(f)
            parser.set_data(data)
            return parser.parse()
        else:
            return None

    def pull_git(self):
        dest_dir = self.get_src_dir()
        git.clone_repo(self.dependency.url, dest_dir, self.dependency.ref)

    def pull_local(self):
        dest_dir = self.get_src_dir()
        if os.path.exists(dest_dir):
            if os.path.isdir(dest_dir):
                shutil.rmtree(dest_dir)
            else:
                os.remove(dest_dir)
        shutil.copytree(self.dependency.url, dest_dir)

    def get_src_dir(self):
        return os.path.join(self.storage.get_sources_dir(), self.dependency.name)

    def get_build_dir(self):
        return os.path.join(self.storage.get_builds_dir(), self.dependency.name)

    def get_buildsystem(self):
        """ Guesses the build system """
        src_path = self.get_src_dir()
        if os.path.isfile(os.path.join(src_path, 'CMakeLists.txt')):
            return BuildSystem.CMAKE
        if os.path.isfile(os.path.join(src_path, 'Makefile')):
            return BuildSystem.MAKEFILE
        return BuildSystem.UNKNOWN

    def build_cmake(self):
        src_dir = self.get_src_dir()
        build_dir = self.get_build_dir()
        install_dir = self.storage.get_install_dir()
        os.makedirs(build_dir, exist_ok=True)

        # Configure
        self.call(
            [
                'cmake',
                '-G', self.options.cmake_generator,
                src_dir,
                '-DCMAKE_INSTALL_PREFIX={0}'.format(install_dir),
                '-DCMAKE_PREFIX_PATH={0}'.format(install_dir)
            ],
            cwd=build_dir
        )

        # Build
        self.call(
            ['cmake', '--build', '.'],
            cwd=build_dir,
        )

        # Install
        self.call(
            ['cmake', '-DCMAKE_INSTALL_CONFIG_NAME=Debug', '-P', 'cmake_install.cmake'],
            cwd=build_dir
        )

    def build_makefile(self):
        self.view.error('Building makefile is not supported yet')
        raise BuildError()

    def call(self, args, cwd):
        self.view.verbose('Calling subprocess: "{0}", cwd: {1}'.format(repr(args), repr(cwd)))
        proc = subprocess.run(args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                              encoding='utf8')

        self.view.verbose('Process output:{0}{1}'.format(os.linesep, str(proc.stdout)))
        if len(proc.stderr) > 0:
            self.view.error(str(proc.stderr))

        if proc.returncode is not 0:
            raise BuildError()
